//=========================================================
// LPC1114 Project
//=========================================================
// File Name : main.c
// Function  : Main Routine
//---------------------------------------------------------
// Rev.01 2011.03.15 Munetomo Maruyama
//---------------------------------------------------------
// Copyright (C) 2010-2011 Munetomo Maruyama
//=========================================================
// ---- License Information -------------------------------
// Anyone can FREELY use this code fully or partially
// under conditions shown below.
// 1. You may use this code only for individual purpose,
//    and educational purpose.
//    Do not use this code for business even if partially.
// 2. You can copy, modify and distribute this code.
// 3. You should remain this header text in your codes
//   including Copyright credit and License Information.
// 4. Your codes should inherit this license information.
//=========================================================
// ---- Patent Notice -------------------------------------
// I have not cared whether this system (hw + sw) causes
// infringement on the patent, copyright, trademark,
// or trade secret rights of others. You have all
// responsibilities for determining if your designs
// and products infringe on the intellectual property
// rights of others, when you use technical information
// included in this system for your business.
//=========================================================
// ---- Disclaimers ---------------------------------------
// The function and reliability of this system are not
// guaranteed. They may cause any damages to loss of
// properties, data, money, profits, life, or business.
// By adopting this system even partially, you assume
// all responsibility for its use.
//=========================================================

#ifdef __USE_CMSIS
#include "LPC11xx.h"
#endif

#include "array_com.h"
#include "color_led.h"
#include "fixedpoint.h"
#include "mems.h"
#include "oled.h"
#include "pwm.h"
#include "systick.h"
#include "utility.h"

//======================
// Break Out Game
//======================
//--------------
// Constants
//--------------
#define LOOP_INTERVAL 1 // 10ms
//
#define BOARD_BLOCK  0 // OB/MB for Block side
#define BOARD_PADDLE 1 // OB/MB for Paddle side
#define BOARD_BEEP   2 // UB/MB for Beep
//
#define SXMIN   0 // screen xmin
#define SYMIN   0 // screen ymin
#define SXMAX 128 // screen xmax
#define SYMAX 128 // screen ymax
#define BALLSIZE  6 // ball size
#define WALLWIDTH 4 // wall width
#define XMIN (WALLWIDTH)                          // field xmin
#define XMAX ((SXMAX) - (WALLWIDTH) - (BALLSIZE)) // field xmax
#define YMIN (WALLWIDTH)                          // field ymin
#define YMAX ((SYMAX) - (WALLWIDTH) - (BALLSIZE)) // field ymax
//
#define BLOCK_SIZEX 15 // block size x
#define BLOCK_SIZEY  8 // block size y
#define BLOCK_COUNTX 8 // how many blocks along x
#define BLOCK_COUNTY 8 // how many blocks along y
#define BLOCK_XMIN (WALLWIDTH)                   // block position xmin
#define BLOCK_YMIN (WALLWIDTH + BLOCK_SIZEY * 4) // block position ymin
#define BLOCK_PATTERNS 4 // how many block patterns
#define BLOCK_COLORS   8 // how many block colors
//
#define PADDLE_SIZEX 30 // paddle size x
#define PADDLE_SIZEY  5 // paddle size y
#define PADDLE_POSY 80  // paddle position y
#define PADDLE_CENTERX ((XMAX + XMIN - PADDLE_SIZEX) >> 1)   // paddle center x
#define PADDLE_XMIN (XMIN)                                   // paddle position xmin
#define PADDLE_XMAX ((SXMAX) - (WALLWIDTH) - (PADDLE_SIZEX)) // paddle position xmax
#define PADDLE_COLOR (OLED_YEL) // paddle color
#define PADDLE_AVE_COUNT 4      // how many accumulate to get averaged paddle position
#define PADDLE_SHOOT_LEVEL 16   // threshold of MEMS value to shoot ball from paddle
#define PADDLE_COUNT 5 // how many paddles do you have first?
//
#define BALL_SPEED_UP 10 // increase ball speed every when paddle hits ball this count
#define BALL_SPEED_LIMIT (BALLSIZE) // ball speed limit
//
#define SCORE_POSY 100 // score display position
//
// Operation State
//
enum OPERATION_STATE {GAME_INIT, GAME_OVER, GAME_START, GAME_RESTART, BALL_ON_PADDLE, BALL_IN_FIELD, BALL_EXIT, BALL_WAIT};
//
// Define Colors
//
static const uint32_t BLK_COLOR[BLOCK_COLORS]
= {OLED_RED, OLED_GRN, OLED_BLU, OLED_WHT, OLED_YEL, OLED_CYN, OLED_MAG, OLED_GRY};
//
// Packet Field
//
#define PACKET_GAME_START 0x80
#define PACKET_BALL_EXIT  0x40
#define PACKET_HIT_WALL   0x20
#define PACKET_HIT_BLOCK  0x10
#define PACKET_HIT_PADDLE 0x08
//
// Block Patterns
//
static const uint32_t BLK_PATTERN[BLOCK_PATTERNS][BLOCK_COUNTY][BLOCK_COUNTX]
= {
      {
          {1, 1, 1, 1, 1, 1, 1, 1},
          {1, 1, 1, 1, 1, 1, 1, 1},
          {1, 1, 1, 1, 1, 1, 1, 1},
          {1, 1, 1, 1, 1, 1, 1, 1},
          {1, 1, 1, 1, 1, 1, 1, 1},
          {1, 1, 1, 1, 1, 1, 1, 1},
          {1, 1, 1, 1, 1, 1, 1, 1},
          {1, 1, 1, 1, 1, 1, 1, 1}
      },
      {
          {1, 1, 1, 1, 1, 1, 1, 1},
          {1, 1, 1, 1, 1, 1, 1, 0},
          {1, 1, 1, 1, 1, 1, 0, 0},
          {1, 1, 1, 1, 1, 0, 0, 0},
          {1, 1, 1, 1, 0, 0, 0, 0},
          {1, 1, 1, 0, 0, 0, 0, 0},
          {1, 1, 0, 0, 0, 0, 0, 0},
          {1, 0, 0, 0, 0, 0, 0, 0}
      },
      {
          {1, 0, 1, 0, 1, 0, 1, 0},
          {0, 1, 0, 1, 0, 1, 0, 1},
          {1, 0, 1, 0, 1, 0, 1, 0},
          {0, 1, 0, 1, 0, 1, 0, 1},
          {1, 0, 1, 0, 1, 0, 1, 0},
          {0, 1, 0, 1, 0, 1, 0, 1},
          {1, 0, 1, 0, 1, 0, 1, 0},
          {0, 1, 0, 1, 0, 1, 0, 1}
      },
      {
          {1, 1, 1, 1, 1, 1, 1, 1},
          {1, 0, 0, 0, 0, 0, 0, 1},
          {1, 0, 1, 1, 1, 1, 0, 1},
          {1, 0, 1, 0, 0, 1, 0, 1},
          {1, 0, 1, 0, 0, 1, 0, 1},
          {1, 0, 1, 1, 1, 1, 0, 1},
          {1, 0, 0, 0, 0, 0, 0, 1},
          {1, 1, 1, 1, 1, 1, 1, 1}
      }
  };

//------------------
// Structure Block
//------------------
typedef struct
{
    int32_t posX_min; // block position xmin
    int32_t posX_max; // block position xmax
    int32_t posY_min; // block position ymin
    int32_t posY_max; // block position ymax
    uint32_t exist;   // block exists?
    uint32_t erase;   // block erased
    uint32_t color;   // block color
} sBlock;
//------------------
// Structure Packet
//------------------
typedef struct
{
    int8_t   packet_info; // Packet Information (see constants "Packet Field")
    int8_t   ballX;  // ball position X
    int8_t   ballY;  // ball position Y
    int8_t   ballDX; // ball speed X
    int8_t   ballDY; // ball speed Y
    uint16_t score;  // score value
} sPacket;

//====================================
// Draw Score
//====================================
void Draw_Score(uint32_t score, uint32_t score_highest, uint32_t paddle_left)
{
    OLED_printf_Font(OLED_FONT_SMALL);
    //
    OLED_printf_Position(1, 14);
    OLED_printf_Color(OLED_RED, OLED_BLK);
    OLED_printf("Breakout!  ");
    OLED_printf_Color(OLED_YEL, OLED_BLK);
    OLED_printf("P=%1d", paddle_left);
    //
    OLED_printf_Position(1, 15);
    OLED_printf_Color(OLED_GRN, OLED_BLK);
    OLED_printf("S=%04d  ", score);
    OLED_printf_Color(OLED_BLU, OLED_BLK);
    OLED_printf("H=%04d", score_highest);
}

//====================================
// Main Operation for BOARD_BLOCK
//====================================
uint32_t Main_Board_Block(uint32_t state)
{
    static int32_t  fieldX_min, fieldX_max, fieldY_min, fieldY_max;
    static int32_t  ballX, ballY;
    static int32_t  ballX_prev, ballY_prev;
    static int32_t  ballDX, ballDY;
    //
    static sBlock   block[BLOCK_COUNTY][BLOCK_COUNTX];
    static uint32_t block_pattern;
    static uint32_t block_count;
    static int32_t  x, y;
    //
    static sPacket  packet;
    static uint8_t  packet_info;
    //
    static uint32_t score;
    //-----------------------
    // State Distribution
    //-----------------------
    switch(state)
    {
        //--------------------
        // Initialize Game
        //--------------------
        case GAME_INIT:
        {
            //
            // Set Field Size
            //
            fieldX_min = XMIN;
            fieldX_max = XMAX;
            fieldY_min = YMIN;
            fieldY_max = SYMAX + BALLSIZE;
            //
            // Next State
            //
            state = GAME_START;
            break;
        }
        //-------------------
        // Start Game
        //-------------------
        case GAME_START:
        {
            score = 0;
            block_pattern = 0;
            //
            // Next State
            //
            state = GAME_RESTART;
            break;
        }
        //-------------------
        // ReStart Game
        //-------------------
        case GAME_RESTART:
        {
            //
            // Draw Wall
            //
            OLED_Fill_Rect(SXMIN, SYMIN,             SXMAX - SXMIN, WALLWIDTH, OLED_YEL); // TOP
            OLED_Fill_Rect(SXMIN, SYMIN,             WALLWIDTH, SYMAX - SYMIN, OLED_YEL); // LEFT
            OLED_Fill_Rect(SXMAX - WALLWIDTH, SYMIN, WALLWIDTH, SYMAX - SYMIN, OLED_YEL); // RIGHT
            //
            // Draw Blocks
            //
            block_count = 0;
            for (y = 0; y < BLOCK_COUNTY; y++)
            {
                for (x = 0; x < BLOCK_COUNTX; x++)
                {
                    block[y][x].exist = BLK_PATTERN[block_pattern][y][x];
                    block[y][x].erase = 0;
                    block[y][x].posY_min  = BLOCK_YMIN + y * BLOCK_SIZEY;
                    block[y][x].posX_min  = BLOCK_XMIN + x * BLOCK_SIZEX;
                    block[y][x].posX_max  = block[y][x].posX_min + BLOCK_SIZEX;
                    block[y][x].posY_max  = block[y][x].posY_min + BLOCK_SIZEY;
                    block[y][x].color = BLK_COLOR[(x + y) % BLOCK_COLORS];
                    //
                    if (block[y][x].exist)
                    {
                        OLED_Fill_Rect(block[y][x].posX_min, block[y][x].posY_min, BLOCK_SIZEX, BLOCK_SIZEY, block[y][x].color);
                        block_count = block_count + 1;
                    }
                    else
                    {
                        OLED_Fill_Rect(block[y][x].posX_min, block[y][x].posY_min, BLOCK_SIZEX, BLOCK_SIZEY, OLED_BLK);
                    }
                }
            }
            //
            // Next State
            //
            state = BALL_WAIT;
            break;
        }
        //--------------------
        // Ball Operation
        //--------------------
        case BALL_IN_FIELD:
        {
            //
            // Initialize Packet Information
            //
            packet_info = 0;
            //
            //
            // Redraw Ball
            //
            if ((ballX_prev != ballX) || (ballY_prev != ballY))
            {
                OLED_Fill_Rect(ballX_prev, ballY_prev, BALLSIZE, BALLSIZE, OLED_BLK);
                OLED_Fill_Rect(ballX,      ballY,      BALLSIZE, BALLSIZE, OLED_CYN);
                ballX_prev = ballX;
                ballY_prev = ballY;
            }
            //
            // Erase Block
            //
            for (y = 0; y < BLOCK_COUNTY; y++)
            {
                for (x = 0; x < BLOCK_COUNTX; x++)
                {
                    if (block[y][x].erase)
                    {
                        block[y][x].exist = 0;
                        block[y][x].erase = 0;
                        score = score + 1;
                        block_count = block_count - 1;
                        packet_info = packet_info | PACKET_HIT_BLOCK;
                        //
                        OLED_Fill_Rect(block[y][x].posX_min, block[y][x].posY_min, BLOCK_SIZEX, BLOCK_SIZEY, OLED_BLK);
                    }
                }
            }
            //
            // Ball Movement
            //
            ballX = ballX + ballDX;
            ballY = ballY + ballDY;
            //
            // Hit to Block?
            //
            for (y = 0; y < BLOCK_COUNTY; y++)
            {
                for (x = 0; x < BLOCK_COUNTX; x++)
                {
                    //
                    // For existing Blocks...
                    //
                    if (block[y][x].exist)
                    {
                        //
                        // Hit on Top Edge?
                        //
                        if ((ballDY > 0)
                         && (ballX >= block[y][x].posX_min - BALLSIZE)
                         && (ballX <  block[y][x].posX_max           )
                         && (ballY >= block[y][x].posY_min - BALLSIZE)
                         && (ballY <  block[y][x].posY_min           ))
                        {
                            ballDY = -ballDY;
                            ballY = block[y][x].posY_min - BALLSIZE;
                            block[y][x].erase = 1;
                        }
                        //
                        // Hit on Bottom Edge?
                        //
                        else if ((ballDY < 0)
                              && (ballX >= block[y][x].posX_min - BALLSIZE)
                              && (ballX <  block[y][x].posX_max           )
                              && (ballY >  block[y][x].posY_max - BALLSIZE)
                              && (ballY <= block[y][x].posY_max           ))
                        {
                            ballDY = -ballDY;
                            ballY = block[y][x].posY_max;
                            block[y][x].erase = 1;
                        }
                        //
                        // Hit on Left Edge?
                        //
                        else if ((ballDX > 0)
                              && (ballX >= block[y][x].posX_min - BALLSIZE)
                              && (ballX <  block[y][x].posX_min           )
                              && (ballY >= block[y][x].posY_min - BALLSIZE)
                              && (ballY <  block[y][x].posY_max           ))
                        {
                            ballDX = -ballDX;
                            ballX = block[y][x].posX_min - BALLSIZE;
                            block[y][x].erase = 1;
                        }
                        //
                        // Hit on Right Edge?
                        //
                        else if ((ballDX < 0)
                              && (ballX >  block[y][x].posX_max - BALLSIZE)
                              && (ballX <= block[y][x].posX_max           )
                              && (ballY >= block[y][x].posY_min - BALLSIZE)
                              && (ballY <  block[y][x].posY_max           ))
                        {
                            ballDX = -ballDX;
                            ballX = block[y][x].posX_max;
                            block[y][x].erase = 1;
                        }
                    }
                } // for x
            } // for y
            //
            // Hit to Screen Edge?
            //
            if (ballX < fieldX_min)
            {
                ballDX = -ballDX;
                ballX = fieldX_min;
                packet_info = packet_info | PACKET_HIT_WALL;
            }
            if (ballX > fieldX_max)
            {
                ballDX = -ballDX;
                ballX = fieldX_max;
                packet_info = packet_info | PACKET_HIT_WALL;
            }
            if (ballY < fieldY_min)
            {
                ballDY = -ballDY;
                ballY = fieldY_min;
                packet_info = packet_info | PACKET_HIT_WALL;
            }
            if ((ballY > fieldY_max) && (ballDY > 0))
            {
                ballY = ballY - fieldY_max;
                //
                // Next State
                //
                state = BALL_EXIT;
            }
            //
            // Send Score and Sound Packet
            //
            if ((state == BALL_IN_FIELD) && (packet_info != 0))
            {
                packet.packet_info = packet_info;
                packet.score  = (uint16_t) score;
                //
                Array_COM_Tx_Multi_Bytes(E, (uint8_t*) &packet, sizeof(packet));
            }
            break;
        }
        //---------------------
        // Waiting a Ball
        //---------------------
        case BALL_WAIT:
        {
            //
            // Receive Packet
            //
            if (Array_COM_Rx_Multi_Bytes(E, (uint8_t*) &packet, 0) > 0)
            {
                packet_info = packet.packet_info;
                //
                // Start Game?
                //
                if (packet_info & PACKET_GAME_START)
                {
                    //
                    // Next State
                    //
                    state = GAME_START;
                }
                //
                // Receive Ball?
                //
                else if (packet_info & PACKET_BALL_EXIT)
                {
                    ballX  = (int32_t) packet.ballX;
                    ballY  = (int32_t) packet.ballY;
                    ballDX = (int32_t) packet.ballDX;
                    ballDY = (int32_t) packet.ballDY;
                    //
                    ballY = ballY + fieldY_max;
                    ballX_prev = ballX;
                    ballY_prev = ballY;
                    //
                    // Next State
                    //
                    state = BALL_IN_FIELD;
                }
            }
            break;
        }
        //------------------------------
        // Ball Exists from BOARD_BLOCK
        //------------------------------
        case BALL_EXIT:
        {
            //
            // Send Packet
            //
            packet_info = packet_info | PACKET_BALL_EXIT;
            packet.packet_info = packet_info;
            packet.ballX  = (int8_t) ((ballX < -128)? -128 : (ballX > 127)? 127 : ballX);
            packet.ballY  = (int8_t) ((ballY < -128)? -128 : (ballY > 127)? 127 : ballY);
            packet.ballDX = (int8_t) ((ballDX < -128)? -128 : (ballDX > 127)? 127 : ballDX);
            packet.ballDY = (int8_t) ((ballDY < -128)? -128 : (ballDY > 127)? 127 : ballDY);
            packet.score  = (uint16_t) score;
            //
            Array_COM_Tx_Multi_Bytes(E, (uint8_t*) &packet, sizeof(packet));
            //
            // Erase Ball
            //
            OLED_Fill_Rect(ballX_prev, ballY_prev, BALLSIZE, BALLSIZE, OLED_BLK); // erase
            //
            // Next State
            //
            state = BALL_WAIT;
            //
            // If all Block are Erased, Make new Block Pattern
            //
            if (block_count == 0)
            {
                block_pattern =  (block_pattern + 1) % BLOCK_PATTERNS;
                //
                // Next State (overwrite)
                //
                state = GAME_RESTART;
            }
            break;
        }
    } // switch
    //----------------
    // Return state
    //----------------
    return state;
}

//====================================
// Main Operation for BOARD_PADDLE
//====================================
uint32_t Main_Board_Paddle(uint32_t state)
{
    static int32_t  fieldX_min, fieldX_max, fieldY_min, fieldY_max;
    static int32_t  ballX, ballY;
    static int32_t  ballX_prev, ballY_prev;
    static int32_t  ballDX, ballDY;
    static int32_t  paddleX;
    static int32_t  paddleX_prev;
    static uint32_t paddle_hit_count;
    static int32_t  memsX;
    static int32_t  memsY;
    static int32_t  memsY_sum;
    static int32_t  memsY_ave;
    static uint32_t memsY_count;
    //
    static sPacket  packet;
    static uint8_t  packet_info;
    //
    static uint32_t score;
    static uint32_t score_highest;
    static uint32_t paddle_left;
    //---------------------------
    // Draw and Move Paddle
    //---------------------------
    if (state != GAME_INIT)
    {
        //
        // Get MEMS Data
        //
        memsY = (int32_t) MEMS_Get_Y(); // X axis of MEMS corresponds to Y axis of Graphics
        memsY_sum = memsY_sum + memsY;
        memsY_count++;
        if (memsY_count >= PADDLE_AVE_COUNT)
        {
            memsY_ave = memsY_sum / PADDLE_AVE_COUNT;
            memsY_count = 0;
            memsY_sum = 0;
        }
        //
        // Redraw Paddle
        //
        paddleX = -memsY_ave * 2 + PADDLE_CENTERX;
        paddleX = (paddleX < PADDLE_XMIN)? PADDLE_XMIN :
                  (paddleX > PADDLE_XMAX)? PADDLE_XMAX : paddleX;
        if (paddleX != paddleX_prev)
        {
            OLED_Fill_Rect(paddleX_prev, PADDLE_POSY, PADDLE_SIZEX, PADDLE_SIZEY, OLED_BLK);
            OLED_Fill_Rect(paddleX,      PADDLE_POSY, PADDLE_SIZEX, PADDLE_SIZEY, PADDLE_COLOR);
            paddleX_prev = paddleX;
        }
    }
    //-----------------------
    // State Distribution
    //-----------------------
    switch(state)
    {
        //--------------------
        // Initialize Game
        //--------------------
        case GAME_INIT:
        {
            //
            // Set Field Size
            //
            fieldX_min = XMIN;
            fieldX_max = XMAX;
            fieldY_min = -BALLSIZE * 2;
            fieldY_max = SCORE_POSY;
            //
            // Set Initial State
            //
            paddleX_prev = PADDLE_CENTERX;
            ballX_prev = paddleX_prev + (PADDLE_SIZEX >> 1) + (BALLSIZE >> 1);
            ballY_prev = PADDLE_POSY - BALLSIZE;
            memsY_sum = 0;
            memsY_count = 0;
            //
            score = 0;
            score_highest = 0;
            paddle_left = PADDLE_COUNT;
            Draw_Score(score, score_highest, paddle_left);
            //
            // Next State
            //
            state = GAME_OVER;
            break;
        }
        //-------------------
        // Game Over
        //-------------------
        case GAME_OVER:
        {
            //
            // Draw Wall
            //
            OLED_Fill_Rect(SXMIN, SYMIN,             WALLWIDTH, SYMAX - SYMIN, OLED_YEL); // LEFT
            OLED_Fill_Rect(SXMAX - WALLWIDTH, SYMIN, WALLWIDTH, SYMAX - SYMIN, OLED_YEL); // RIGHT
            //
            // Draw Game Title
            //
            OLED_printf_Font(OLED_FONT_MEDIUM);
            OLED_printf_Position(1, 0);
            OLED_printf_Color(OLED_WHT, OLED_BLK); OLED_printf("Mary");
            //
            OLED_printf_Font(OLED_FONT_SMALL);
            OLED_printf_Position(2, 3);
            OLED_printf_Color(OLED_RED, OLED_BLK); OLED_printf("B");
            OLED_printf_Color(OLED_GRN, OLED_BLK); OLED_printf("r");
            OLED_printf_Color(OLED_BLU, OLED_BLK); OLED_printf("e");
            OLED_printf_Color(OLED_WHT, OLED_BLK); OLED_printf("a");
            OLED_printf_Color(OLED_YEL, OLED_BLK); OLED_printf("k");
            OLED_printf_Color(OLED_CYN, OLED_BLK); OLED_printf("o");
            OLED_printf_Color(OLED_MAG, OLED_BLK); OLED_printf("u");
            OLED_printf_Color(OLED_GRY, OLED_BLK); OLED_printf("t ");
            OLED_printf_Color(OLED_RED, OLED_BLK); OLED_printf("G");
            OLED_printf_Color(OLED_GRN, OLED_BLK); OLED_printf("a");
            OLED_printf_Color(OLED_BLU, OLED_BLK); OLED_printf("m");
            OLED_printf_Color(OLED_CYN, OLED_BLK); OLED_printf("e");
            //
            // ReStart Game?
            //
            memsX = (int32_t) MEMS_Get_X(); // X axis of MEMS corresponds to Y axis of Graphics
            if (memsX < -30)
            {
                //
                // Erase Game Title
                //
                OLED_printf_Font(OLED_FONT_MEDIUM);
                OLED_printf_Position(1, 0);
                OLED_printf_Color(OLED_WHT, OLED_BLK);
                OLED_printf("    ");
                //
                OLED_printf_Font(OLED_FONT_SMALL);
                OLED_printf_Position(2, 3);
                OLED_printf_Color(OLED_WHT, OLED_BLK);
                OLED_printf("             ");
                //
                score = 0;
                paddle_left = PADDLE_COUNT;
                Draw_Score(score, score_highest, paddle_left);
                //
                // Send Game Start Packet
                //
                packet_info = PACKET_GAME_START;
                packet.packet_info = packet_info;
                Array_COM_Tx_Multi_Bytes(W, (uint8_t*) &packet, sizeof(packet));
                //
                // Next State
                //
                state = BALL_ON_PADDLE;
            }
            break;
        }
        //--------------------
        // Ball on Paddle
        //--------------------
        case BALL_ON_PADDLE:
        {
            //
            // Draw Ball on Paddle
            //
            ballX = paddleX + (PADDLE_SIZEX >> 1) - (BALLSIZE >> 1);
            ballY = PADDLE_POSY - BALLSIZE;
            if ((ballX_prev != ballX) || (ballY_prev != ballY))
            {
                OLED_Fill_Rect(ballX_prev, ballY_prev, BALLSIZE, BALLSIZE, OLED_BLK);
                OLED_Fill_Rect(ballX,      ballY,      BALLSIZE, BALLSIZE, OLED_CYN);
                ballX_prev = ballX;
                ballY_prev = ballY;
            }
            //
            // Shoot Ball from Paddle
            //
            memsX = (int32_t) MEMS_Get_X(); // X axis of MEMS corresponds to Y axis of Graphics
            memsY = (int32_t) MEMS_Get_Y(); // Y axis of MEMS corresponds to X axis of Graphics
            //
            if (memsX > PADDLE_SHOOT_LEVEL) // positive
            {
                paddle_hit_count = 0;
                //
                ballDX = -memsY / PADDLE_SHOOT_LEVEL;
                ballDX = (ballDX < -BALL_SPEED_LIMIT)? -BALL_SPEED_LIMIT : ballDX; // saturate
                ballDX = (ballDX >  BALL_SPEED_LIMIT)?  BALL_SPEED_LIMIT : ballDX; // saturate
                //
                ballDY = -memsX / PADDLE_SHOOT_LEVEL * 2; // negative
                ballDY = (ballDY < -BALL_SPEED_LIMIT)? -BALL_SPEED_LIMIT : ballDY; // saturate
                //
                // Next State
                //
                state = BALL_IN_FIELD;
            }
            break;
        }
        //--------------------
        // Ball Operation
        //--------------------
        case BALL_IN_FIELD:
        {
            //
            // Initialize Packet Information
            //
            packet_info = 0;
            //
            // Redraw Ball
            //
            if ((ballX_prev != ballX) || (ballY_prev != ballY))
            {
                OLED_Fill_Rect(ballX_prev, ballY_prev, BALLSIZE, BALLSIZE, OLED_BLK);
                OLED_Fill_Rect(ballX,      ballY,      BALLSIZE, BALLSIZE, OLED_CYN);
                ballX_prev = ballX;
                ballY_prev = ballY;
            }
            //
            // Ball Movement
            //
            ballX = ballX + ballDX;
            ballY = ballY + ballDY;
            //
            // Hit to Paddle?
            //
            if ((ballY >= (PADDLE_POSY - BALLSIZE)) && (ballY < PADDLE_POSY))
            {
                if ((ballX > (paddleX - BALLSIZE)) && (ballX < (paddleX + PADDLE_SIZEX)))
                {
                    //
                    // Set Ball Direction
                    //
                    ballDX = (ballX - (paddleX - BALLSIZE) - ((PADDLE_SIZEX + BALLSIZE) / 2)) / 4;
                    ballDY = -ballDY;
                    //
                    ballY = PADDLE_POSY- BALLSIZE;
                    packet_info = packet_info | PACKET_HIT_PADDLE;
                    //
                    // Speed Up Ball
                    //
                    paddle_hit_count = paddle_hit_count + 1;
                    if (paddle_hit_count == BALL_SPEED_UP)
                    {
                        paddle_hit_count = 0;
                        if (ballDX > 0) ballDX = ballDX + 1;
                        if (ballDX < 0) ballDX = ballDX - 1;
                        if (ballDY > 0) ballDY = ballDY + 1;
                        if (ballDY < 0) ballDY = ballDY - 1;
                        ballDX = (ballDX < -BALL_SPEED_LIMIT)? -BALL_SPEED_LIMIT : ballDX; // saturate
                        ballDX = (ballDX >  BALL_SPEED_LIMIT)?  BALL_SPEED_LIMIT : ballDX; // saturate
                        ballDY = (ballDY < -BALL_SPEED_LIMIT)? -BALL_SPEED_LIMIT : ballDY; // saturate
                        ballDY = (ballDY >  BALL_SPEED_LIMIT)?  BALL_SPEED_LIMIT : ballDY; // saturate
                    }
                }
            }
            //
            // Hit to Screen Edge?
            //
            if (ballX < fieldX_min)
            {
                ballDX = -ballDX;
                ballX = fieldX_min;
                packet_info = packet_info | PACKET_HIT_WALL;
            }
            if (ballX > fieldX_max)
            {
                ballDX = -ballDX;
                ballX = fieldX_max;
                packet_info = packet_info | PACKET_HIT_WALL;
            }
            //
            if ((ballY < fieldY_min) && (ballDY < 0))
            {
                //
                // Next State
                //
                state = BALL_EXIT;
            }
            //
            if (ballY > fieldY_max)
            {
                //
                // Miss
                //
                paddle_left = paddle_left - 1;
                Draw_Score(score, score_highest, paddle_left);
                //
                // Erase ball
                //
                OLED_Fill_Rect(ballX_prev, ballY_prev, BALLSIZE, BALLSIZE, OLED_BLK);
                //
                // Next State
                //
                state = (paddle_left == 0)? GAME_OVER : BALL_ON_PADDLE;
            }
            //
            // Send Packet to BOARD_BEEP
            //
            if (Array_COM_Port_Open(E))
            {
                if (packet_info & (PACKET_HIT_WALL | PACKET_HIT_PADDLE))
                {
                    packet.packet_info = packet_info;
                    Array_COM_Tx_Multi_Bytes(E, (uint8_t*) &packet, sizeof(packet));
                }
            }
            break;
        }
        //---------------------
        // Waiting a Ball
        //---------------------
        case BALL_WAIT:
        {
            //
            // Receive Packet
            //
            if (Array_COM_Rx_Multi_Bytes(W, (uint8_t*) &packet, 0) > 0)
            {
                packet_info = packet.packet_info;
                //
                // Receive Ball
                //
                if (packet_info & PACKET_BALL_EXIT)
                {
                    ballX  = (int32_t) packet.ballX;
                    ballY  = (int32_t) packet.ballY;
                    ballDX = (int32_t) packet.ballDX;
                    ballDY = (int32_t) packet.ballDY;
                    //
                    ballY = ballY + fieldY_min;
                    ballX_prev = ballX;
                    ballY_prev = ballY;
                    //
                    // Next State
                    //
                    state = BALL_IN_FIELD;
                }
                //
                // Get Score
                //
                score  = (uint32_t) packet.score;
                score_highest = (score > score_highest)? score : score_highest;
                Draw_Score(score, score_highest, paddle_left);
                //
                // Send Packet to BOARD_BEEP
                //
                if (Array_COM_Port_Open(E))
                {
                    if (packet_info & (PACKET_HIT_WALL | PACKET_HIT_BLOCK))
                    {
                        packet.packet_info = packet_info;
                        Array_COM_Tx_Multi_Bytes(E, (uint8_t*) &packet, sizeof(packet));
                    }
                }
            }
            break;
        }
        //------------------------------
        // Ball Exists from BOARD_PADDLE
        //------------------------------
        case BALL_EXIT:
        {
            //
            // Send Ball Packet
            //
            ballY = ballY - fieldY_min;
            packet_info = PACKET_BALL_EXIT;
            //
            packet.packet_info = packet_info;
            packet.ballX  = (int8_t) ((ballX < -128)? -128 : (ballX > 127)? 127 : ballX);
            packet.ballY  = (int8_t) ((ballY < -128)? -128 : (ballY > 127)? 127 : ballY);
            packet.ballDX = (int8_t) ((ballDX < -128)? -128 : (ballDX > 127)? 127 : ballDX);
            packet.ballDY = (int8_t) ((ballDY < -128)? -128 : (ballDY > 127)? 127 : ballDY);
            Array_COM_Tx_Multi_Bytes(W, (uint8_t*) &packet, sizeof(packet));
            //
            OLED_Fill_Rect(ballX_prev, ballY_prev, BALLSIZE, BALLSIZE, OLED_BLK); // erase
            //
            // Next State
            //
            state = BALL_WAIT;
            break;
        }
    } // switch
    //----------------
    // Return state
    //----------------
    return state;
}

//====================================
// Main Operation for BOARD_BEEP
//====================================
uint32_t Main_Board_Beep(uint32_t state)
{
    static sPacket  packet;
    static uint32_t beep;
    //---------------------------------
    // Receive Packet from BOARD_PADDLE
    //---------------------------------
    if (Array_COM_Port_Open(W))
    {
        if (Array_COM_Rx_Multi_Bytes(W, (uint8_t*) &packet, 0) > 0)
        {
            beep = packet.packet_info;
            //
            if (beep & PACKET_HIT_PADDLE)
            {
                Play_Sound(523, 1);
            }
            if (beep & PACKET_HIT_WALL)
            {
                Play_Sound(659, 1);
            }
            if (beep & PACKET_HIT_BLOCK)
            {
                Play_Sound(784, 1);
            }
        }
    }
    //----------------
    // Return state
    //----------------
    return state;
}

//-----------------------
// Main Routine
//-----------------------
int main(void)
{
    uint32_t tick_curr;
    uint32_t tick_prev;
    uint32_t tick_intv;
    uint32_t state;
    //---------------------
    // Initialize Hardware
    //---------------------
    Init_SysTick();
    Init_Array_COM();
    Array_COM_ID_Assignment();
    Init_Color_LED();
    //----------------
    // Initialize OLED
    //----------------
    if ((Array_COM_Get_ID() == BOARD_BLOCK) || (Array_COM_Get_ID() == BOARD_PADDLE))
    {
        Init_OLED_with_Orientation(OLED_TOP_W);
        Init_MEMS();
        OLED_Clear_Screen(OLED_BLK); // Clear Screen
    }
    //----------------
    // Initialize Beep
    //----------------
    if (Array_COM_Get_ID() == BOARD_BEEP)
    {
        Init_PWM();
    }
    //------------------------
    // Check Board Connection
    //------------------------
    if (Array_COM_Get_ID() == BOARD_BLOCK)
    {
        if (Array_COM_Port_Open(E) == 0)
        {
            OLED_printf("Connection Error");
            while(1); // stop
        }
    }
    if (Array_COM_Get_ID() == BOARD_PADDLE)
    {
        if (Array_COM_Port_Open(W) == 0)
        {
            OLED_printf("Connection Error");
            while(1); // stop
        }
    }
    //----------------
    // Initial State
    //----------------
    state = GAME_INIT;
    //===================
    // Main Loop
    //===================
    while(1)
    {
        //-------------------
        // Adjust Loop Timing
        //-------------------
        while(1)
        {
            tick_curr = Get_Ticks();
            tick_intv = tick_curr - tick_prev;
            if (tick_intv > LOOP_INTERVAL) break;
        }
        tick_prev = tick_curr;
        //===============================
        // Main Operation in BOARD_BLOCK
        //===============================
        if (Array_COM_Get_ID() == BOARD_BLOCK)
        {
            state = Main_Board_Block(state);
        }
        //===============================
        // Main Operation in BOARD_PADDLE
        //===============================
        if (Array_COM_Get_ID() == BOARD_PADDLE)
        {
            state = Main_Board_Paddle(state);
        }
        //===============================
        // Main Operation in BOARD_BEEP
        //===============================
        if (Array_COM_Get_ID() == BOARD_BEEP)
        {
            state = Main_Board_Beep(state);
        }
    } // while(1)
    return 0;
}

//=========================================================
// End of Program
//=========================================================
